


#include <cstdio>
#include<math.h>
#include<stdio.h>

#include "nbody.h"

/// kernel routines
#include <nbody_kernel.cu> // included here! is not compiled separately!
 /// this one does kick step and all pairs forces
#include <energy_kernel.cu> 
 /// all pairs potential energy and kinetic energy computation
#include <kep_kernel.cu> 
 /// f+g function keplerian evolution on GPU for N particles
#include <drift_kernel.cu> 
 /// drift step computation on GPU
 /// also contains parallel reduction sum ]
 /// and m*v vector on GPU routine


extern "C"
{

////////////////////////////////////////
/// looks to see if device available and can run CUDA
void checkCUDA(int device)
{   
    cudaError_t err;
    err= cudaSetDevice(device);
    if (err != cudaSuccess) 
         printf("cudaSetDevice fail, device %d\n",device);
    if (err == cudaErrorInvalidDevice)
       printf("cudaErrorInvalidDevice %d\n",device);
//    if (err == cudaErrorSetOnActiveProcess)
//     printf("cudaErrorSetOnActiveProcess\n");
    printf("cudaSetDevice done\n");
}

// never used because last time I tried it
// the device only got 0
void setDeviceSoftening(double softening)
{
    double softeningSq = softening*softening;

    cudaError_t err;
    err = cudaMemcpyToSymbol("softeningSquared", &softeningSq, 
                                      sizeof(double), 0, 
                                      cudaMemcpyHostToDevice);
    if (err != cudaSuccess) 
         printf("cudaMemcpytoSymbol fail in setDeviceSoftening \n");
    printf("softeningSq=%f \n",softeningSq); // this works fine here
}

////////////////////////////////////////
/// allocate memory for nbody system
/// returns a 1 if fails 0 otherwise
/// this trap is because if your GPU is jammed this routine will fail first
int allocateNBodyArrays(double* vel[2], int numBodies)
{
    int rerr=0;
    // 4 doubles each for alignment reasons
    unsigned int memSize = sizeof( double) * 4 * numBodies;
    
    cudaError_t err;
    err = cudaMalloc((void**)&vel[0], memSize);
    if (err != cudaSuccess) {
         printf("cudaMalloc fail [0] in allocateNbodyArrays \n");
         rerr=1;
    }
    err = cudaMalloc((void**)&vel[1], memSize);
    if (err != cudaSuccess) {
         printf("cudaMalloc fail [1] in allocateNbodyArrays \n");
         rerr=1;
    }
    
    return rerr;
}


////////////////////////////////////////
/// cleaup
void deleteNBodyArrays(double* vel[2])
{
    cudaError_t err;
    err = cudaFree((void**)vel[0]);
    if (err != cudaSuccess) 
         printf("cudaFree fail [0] in deleteNbodyArrays \n");
    err = cudaFree((void**)vel[1]);
    if (err != cudaSuccess) 
         printf("cudaFree fail [1] in deleteNbodyArrays \n");
}

////////////////////////////////////////
// note times 4 in size
void copyArrayFromDevice(double* host, 
                         const double* device, 
                         int numBodies)
{   
    cudaError_t err;
    err = cudaMemcpy(host, device, numBodies*4*sizeof(double), 
                              cudaMemcpyDeviceToHost);
    if (err != cudaSuccess) 
         printf("cudaMemcpy fail in copyArrayFromDevice \n");
}

////////////////////////////////////////
// note times 4 in size
void copyArrayToDevice(double* device, const double* host, int numBodies)
{
    cudaError_t err;
    err = cudaMemcpy(device, host, numBodies*4*sizeof(double), 
                              cudaMemcpyHostToDevice);
    if (err != cudaSuccess) 
         printf("cudaMemcpy fail in copyArrayToDevice \n");
}
// note newPos and other arrays are recast here in this routine
// from double* to double4* 


///////////////////////////////////////////////////////
/// change the central mass in global device memory
/// primarily used during energy computation
/// other routines assume central mass set to zero previously
///////////////////////////////////////////////////////
void
callchangem0(double *aPos, double m0) 
{
    double Pos0[4]={0.0,0.0,0.0,0.0};
    Pos0[3]  = m0;
    cudaError_t err;
    err = cudaMemcpy(aPos,Pos0,sizeof(double)*4, cudaMemcpyHostToDevice); 
    if (err != cudaSuccess) 
         printf("cudaMemcpy fail in callchangem0 \n");
}

// forward declare
void sumMomenta(double* Pos, double* Vel, int p, int numBodies, double *pxpypz);

////////////////////////////////////////////////////////////
/** compute the energyvector and then sum it
   if m0==0 then cartesian coordinates assumed
   if m0>0 then helio/bary coordinates assumed
   energyvector all pairs computed on GPU
   sums done with parallel reduction algorith
   note that energyvector done with m0!=0 if helio/bary
   V0 = Q0 = 0 should be set previously
   if massless particles included later on, then numBodiesMassive 
   is only those massive because only they contribute to energy
   numBodiesMassive is forced to be an integer multiple of pthreads
*/
////////////////////////////////////////////////////////////
double energysum(double* Pos,  double* Vel,
                     double soft2, int numBodiesMassive, int p, double m0)
{

    int npmassive = (int)ceil((float)numBodiesMassive/(float)p);
    npmassive *= p;   /// this is now a multiple of pthreads

    int sharedMemSize = p *  sizeof(double4); // 4 doubles for pos
    dim3 threads(p,1,1); /// number of threads per block is p  blockdim.x=p
    dim3 grid(npmassive/p, 1, 1);

    double *d_vec; /// energy vector on device
    cudaError_t err;
    err = cudaMalloc( (void**) &d_vec, npmassive*sizeof(double));
    if (err != cudaSuccess) printf("cudaMalloc fail d_vec in energysum\n");

    if (m0 >0.0) callchangem0(Pos, m0);  

/**   if helio/bary then reset central mass to non zero
      this is so central mass interaction terms are added
      as long as Q0 com term = 0 then the central mass interaction
      terms should be correct
      if we have set V0 to zero then it doesn't matter
      that KE added in 0 term also
*/

/// computes a vector that has 0.5* Sum_i m_i *[ v_i^2 - Sum_{j!=i} m_j/r_ij ]
    BodyKEPE<<< grid, threads, sharedMemSize >>>
        (d_vec, (double4*) Pos, (double4*) Vel, soft2, npmassive);
    /// d_vec is now the energy vector on device
    err = cudaThreadSynchronize();
    if (err != cudaSuccess) printf("BodyKEPE kernel execution failed\n");

/// we speed this up with reduction parallel sum
/// for parallel reduction sum kernel
    int phalf=p;
    if (npmassive%(2*p) != 0 ) phalf = p/2; 
    dim3 threads2(phalf,1,1); 
    int sharedMemSize2 = phalf*sizeof(double); /// needed for parallel reduction sum
    int numBlocks = npmassive/(2*phalf); // note 2
    numBlocks*=2;   // reduce6 is meant to run with more blocks
    dim3 grid2(numBlocks, 1, 1);

/// this may solve the problem of needing numBodies to be a factor of 2p

    double *d_odata;   /// for sums on device
    err = cudaMalloc((void**) &d_odata, numBlocks*sizeof(double));
    if (err != cudaSuccess) printf("cudaMalloc fail d_odata in energysum\n");
    double *h_odata = (double *) malloc(numBlocks*sizeof(double));    // sums on CPU

    /// do the reduction sum each block, switch on blockSize 
    switch(phalf) { // may need more cases here 
        case  32:
         reduce6<32> <<< grid2, threads2, sharedMemSize2>>>
                 (d_vec, d_odata, npmassive);
         break;
        case  64:
         reduce6<64> <<< grid2, threads2, sharedMemSize2>>>
                 (d_vec, d_odata, npmassive);
         break;
        case 128:
         reduce6<128> <<< grid2, threads2, sharedMemSize2>>>
                 (d_vec, d_odata, npmassive);
         break;
        case 256:
         reduce6<256> <<< grid2, threads2, sharedMemSize2>>>
                 (d_vec, d_odata, npmassive);
         break;
        case 512:
         reduce6<512> <<< grid2, threads2, sharedMemSize2>>>
                 (d_vec, d_odata, npmassive);
         break;
    }
    err = cudaThreadSynchronize();
    if (err != cudaSuccess) printf("reduce6 kernel execution failed\n");

    err = cudaMemcpy( h_odata, d_odata, numBlocks*sizeof(double),
       cudaMemcpyDeviceToHost);
    if (err != cudaSuccess) printf("cudaMemcpy fail in energysum\n");
    double gpu_result = 0.0;
    /// sum the block sums
    for(int i=0; i<numBlocks; i++) gpu_result += h_odata[i];

/// cleanup so far
    err = cudaFree(d_vec);
    if (err != cudaSuccess) printf("cudaFree fail in energysum\n");
    err = cudaFree(d_odata);
    if (err != cudaSuccess) printf("cudaFree fail in energysum\n");
    free(h_odata);

/// if bary/helio extra momentum term needed 
    if (m0 >0.0){  
      callchangem0(Pos, 0.0);  // reset central mass back to zero
      double pxpypz[3] = {0.0,0.0,0.0};
      // call sum momenta routine
      sumMomenta(Pos,  Vel, p, npmassive, pxpypz);// returns sum of P
      double p2sum = pxpypz[0]*pxpypz[0]+pxpypz[1]*pxpypz[1]+pxpypz[2]*pxpypz[2];
      gpu_result += 1.0/(2.0*m0)*p2sum;  // add in the last needed term
    }
    
    return gpu_result;
}


///////////////////////////////////////////////////////
/** Take a kepstep for all bodies, calls fg function routine on GPU
   using universal coordinate differential Kepler eqn solver
   assumes already in helio/bary coordinates
   does not care about central mass
   for all bodies, massive + massless
*/
///////////////////////////////////////////////////////
void
kepstepNbodySystem(double* newPos, double* newVel,
                     double* oldPos, double* oldVel,
                     double deltaTime, 
                     int numBodies, int p, double m0)
{
    int sharedMemSize = 0; // none needed for a kepstep
    dim3 threads(p,1,1); // number of threads per block is p  blockdim.x=p
    dim3 grid(numBodies/p, 1, 1);
    // number of blocks is griddim.x=numBodies/p and gridDim.y=1
    kepstepBodies<<< grid, threads, sharedMemSize >>>
            ((double4*)newPos, (double4*)newVel,
            (double4*)oldPos, (double4*)oldVel,
            deltaTime,  numBodies, m0);

    cudaError_t err = cudaThreadSynchronize();
    if (err != cudaSuccess) printf("KepstepBodies kernel execution failed\n");

}

///////////////////////////////////////////////////////
/** sum m*Vx m*Vy and m*Vy for all particles on GPU
   returned in pxpypz vector on host
   Pos and Vel are on device
   call GPU for m*v sum, not necessarily slow because in global mem
   call GPU for parallel reduction sum
   note central mass or V0 should be set to zero 
   before this is called
   This routine is needed for energy sum and for drift term computation
   can be all particles massive+massless though massless don't contribute
   This routine might screw up if numBodies is not divisible by 2*p
*/
///////////////////////////////////////////////////////
void sumMomenta(double* Pos, double* Vel, int p, int numBodies, double *pxpypz)
{
// for sum vector kernel
    int sharedMemSize = 0; // none needed for the sum vector
    dim3 threads(p,1,1); // number of threads per block is p  blockdim.x=p
    dim3 grid(numBodies/p, 1, 1);
    cudaError_t err;

// for parallel reduction sum kernel
    int phalf=p;
    if (numBodies%(2*p) != 0 ) phalf = p/2; 
// this may solve the problem of needing numBodies to be a factor of 2p
     
    dim3 threads2(phalf,1,1); 
    int sharedMemSize2 = phalf*sizeof(double); 
            // needed for parallel reduction sum 
    int numBlocks = numBodies/(2*phalf); // note 2
    numBlocks*=2;   // reduce6 is meant to run with more blocks
    dim3 grid2(numBlocks, 1, 1);

    double* d_mv; // allocate mv vector on device
    err = cudaMalloc((void**) &d_mv, numBodies*sizeof(double)) ;
    if (err != cudaSuccess) printf("cudaMalloc fail d_mv in sumMomenta \n");
    double *d_odata;   // for sums on device
    err = cudaMalloc((void**) &d_odata, numBlocks*sizeof(double)) ;
    if (err != cudaSuccess) printf("cudaMalloc fail d_odata in sumMomenta \n");
    double *h_odata = (double *) malloc(numBlocks*sizeof(double));    // sums on CPU

// for x, y, z
    for(int coordpos=0;coordpos<3;coordpos++)
    {
      // compute the m*v vector
      computemv <<< grid, threads, sharedMemSize>>>(Pos, Vel, d_mv, coordpos);
      err = cudaThreadSynchronize();
      if (err != cudaSuccess) printf("computemv kernel execution failed\n");

      // do the reduction sum each block, switch on blockSize
      switch(phalf) { // may need more cases here 
        case  32:
         reduce6<32> <<< grid2, threads2, sharedMemSize2>>>(d_mv, d_odata, numBodies);
         break;
        case  64:
         reduce6<64> <<< grid2, threads2, sharedMemSize2>>>(d_mv, d_odata, numBodies);
         break;
        case 128:
         reduce6<128> <<< grid2, threads2, sharedMemSize2>>>(d_mv, d_odata, numBodies);
         break;
        case 256:
         reduce6<256> <<< grid2, threads2, sharedMemSize2>>>(d_mv, d_odata, numBodies);
         break;
        case 512:
         reduce6<512> <<< grid2, threads2, sharedMemSize2>>>(d_mv, d_odata, numBodies);
         break;
      }
      err = cudaThreadSynchronize();
      if (err != cudaSuccess) printf("reduce6 kernel execution failed\n");
   
      err = cudaMemcpy( h_odata, d_odata, numBlocks*sizeof(double), 
         cudaMemcpyDeviceToHost);
      if (err != cudaSuccess) printf("cudaMemcpy fail in sumMomenta \n");
    
      double gpu_result = 0.0;
      // sum the block sums
      for(int i=0; i<numBlocks; i++) gpu_result += h_odata[i];
      pxpypz[coordpos] = gpu_result;  // return the result in a vector

    }

// cleanup
    err = cudaFree(d_odata);
    if (err != cudaSuccess) printf("cudaFree fail in sumMomenta \n");
    err = cudaFree(d_mv);
    if (err != cudaSuccess) printf("cudaFree fail in sumMomenta \n");
    free(h_odata);

}

///////////////////////////////////////////////////////
/** do interaction step for all bodies
   note central mass should be set to zero before this routine is called
   velocity kicks for all particles
   all pair forces used here
   If there are massless particles:
      Then the grid should be set up for all particles
      but interactionStep should be called with numBodiesMassive 
      both should have p=nthreads as factors
      this should be true:  numBodies > numBodiesMassive
*/
///////////////////////////////////////////////////////
void
kickstepNbodySystem(double* newPos, double* newVel,
                     double* oldPos, double* oldVel,
                     double deltaTime, double damping,  double soft2,
                     int numBodies, int numBodiesMassive, int p, double m0)
{
    static int firsttime = 0;
    int sharedMemSize = p *  sizeof(double4); // 4 doubles for pos
    dim3 threads(p,1,1); // number of threads per block is p  blockdim.x=p
    dim3 grid(numBodies/p, 1, 1);
    callchangem0(oldPos, 0.0);   // make sure central mass is zero here!

    int npmassive = (int)ceil((float)numBodiesMassive/(float)p);
    npmassive *= p;   // this is now a multiple of pthreads
    
    // number of blocks is griddim.x=numBodies/p and gridDim.y=1
    if (firsttime ==0){
       printf("number blocks %d, number threads=%d\n",grid.x,threads.x);
       fflush(stdout);
       firsttime=1;
    }

    // execute the interaction kernel: type conversion to double4 done here
    
    cudaError_t err = cudaThreadSynchronize();
    interactionStep<<< grid, threads, sharedMemSize >>>
            ((double4*)newPos, (double4*)newVel,
            (double4*)oldPos, (double4*)oldVel,
            deltaTime, damping, soft2, npmassive);

    // check if kernel invocation generated an error
    err = cudaThreadSynchronize();
    if (err != cudaSuccess) printf("interactionStep kernel execution failed\n");
}


///////////////////////////////////////////////////////
/** Hsun drift term done for all bodies
   can be called with deltatimefac = deltatime/(2*m0) for a step tau/2
   will work for all particles massive+massless
   should be called for all bodies as they need to drift too
   first mass should be 0
*/
///////////////////////////////////////////////////////
void  
driftstepNbodySystem(double* newPos, double* newVel,
                     double* oldPos, double* oldVel,
                     double deltaTimefac, 
                     int numBodies, int p)
{
    // get sum of mv of all particles GPU routine 
    double pxpypz[3] = {0.0,0.0,0.0};
    callchangem0(oldPos, 0.0);  // set central mass zero here
    sumMomenta(oldPos, oldVel, p, numBodies, pxpypz);
    
/*
    double px=0.0; double py=0.0; double pz=0.0;
    for(int i=0;i<numBodies;i++){
      px += oldPos[4*i+3]*oldVel[4*i+0];
      py += oldPos[4*i+3]*oldVel[4*i+1];
      pz += oldPos[4*i+3]*oldVel[4*i+2];
    }
    printf("pxpypxcpu = %.3e %.3e %.3e\n",px,py,pz);
    printf("pxpypxgpu = %.3e %.3e %.3e\n",pxpypz[0],pxpypz[1],pxpypz[2]);
*/

    int sharedMemSize = 0; // none needed for a driftstep
    dim3 threads(p,1,1); 
    dim3 grid(numBodies/p, 1, 1);

    cudaError_t err;
    double *d_SumP;  // on device
    err =  cudaMalloc((void**) &d_SumP, 3*sizeof(double)) ;
    if (err != cudaSuccess) printf("cudaMalloc fail in driftstepNbodySystem\n");
    err =  cudaMemcpy( d_SumP, pxpypz, 3*sizeof(double),
        cudaMemcpyHostToDevice) ;
    if (err != cudaSuccess) printf("cudaMemcpy fail in driftstepNbodySystem\n");
   
    driftstepBodies<<< grid, threads, sharedMemSize >>>
            ((double4*)newPos, (double4*)newVel,
            (double4*)oldPos, (double4*)oldVel,
            deltaTimefac,  (double3*)d_SumP);

    err = cudaThreadSynchronize();
    if (err != cudaSuccess) printf("driftstepBodies execution failed\n");
    err = cudaFree(d_SumP);
    if (err != cudaSuccess) printf("cudaFree fail in driftstepNbodySystem\n");

}


// I did try halving number of threads in Kepstep 
// in case e register space is an issue, but was not necessary
#define KEPFAC 2   
//  divide threads for that routine
///////////////////////////////////////////////////////
/** single timestep done with all evolution operators
   a is out or new, b is old positions and velocities
   algorithm outlined by Duncan et al. with all 3 types of steps
   here m0 is GM
*/
///////////////////////////////////////////////////////
void 
integrateKepNbodySystem(double* aPos, double* aVel,   
                     double* bPos, double* bVel, 
                     double deltaTime, double damping,  double soft2,
                     int numBodies, int numBodiesMassive, int p, double m0)
{

  kepstepNbodySystem(aPos, aVel, bPos, bVel,
           deltaTime/2.0, numBodies, p/KEPFAC, m0); 

  driftstepNbodySystem(bPos, bVel, aPos, aVel,
           deltaTime/(2.0*m0), numBodies, p);

  kickstepNbodySystem(aPos, aVel, bPos, bVel,
           deltaTime, damping,  soft2, numBodies, numBodiesMassive, p,m0);

  driftstepNbodySystem(bPos, bVel, aPos, aVel,
           deltaTime/(2.0*m0), numBodies, p);

  kepstepNbodySystem(aPos, aVel, bPos, bVel,
           deltaTime/2.0, numBodies, p/KEPFAC, m0);

}

}
