

#include<cstdio>
#include<cstdlib>
#include<math.h>

// coordinate transformation routines

/* routines:
correctcom: correct to center of mass
tobary      transfer to barycentric momenta and heliocentric coordinates
frombary    reverse transfer
energycart  compute energy for cartesian coords
*/

// some errors fixed in frombary Oct 20 2009 by AQ
// also simplified at same time
     
//////////////////////////////////////////////////
// correct for center of mass and center of momentum 
// i.e. subtract position and velocity of com
// assumed cartesian coordinates here
void correctcom(double *hPos, double *hVel, int numBodies)
{
   double sum[3] = {0.0,0.0,0.0};  // sum of m*xi 
   double sumv[3] = {0.0,0.0,0.0}; // sum of m*vi
   double mtot=0.0;
   int j;
   for (int i=0;i<numBodies;i++){
      j=4*i;
      double mi = hPos[j+3];
      mtot += mi;
      for(int k=0;k<3;k++){
        sum[k] += mi*hPos[j+k]; sumv[k]+= mi*hVel[j+k]; 
      }
   }
   for(int k=0;k<3;k++){
     sum[k] /= mtot; sumv[k]/= mtot;
   }
   for (int i=0;i<numBodies;i++){
     j=4*i;
     for(int k=0;k<3;k++){
       hPos[j+k] -= sum[k];  hVel[j+k] -= sumv[k]; 
     }
   }

}

//////////////////////////////////////////////////
// convert cartesian to barycenter momenta and helio positions
// using equation 27 and 30 by Duncan et al. 98
// The citation is Duncan, M. J. Levison, H. F., \& Lee, M. H.
// Astronomical Journal, 116, 2067-2077, 1998
// "A Multiple Time Step Symplectic Algorithm for Integrating Close Cncounters"
//  hVelhel, hPoshel are outputs
//  hPos, hVel are inputs 
// note: zero position mass remains m0 and does not become mtot
// note: V_i = P_i/m_i but is not equal to \dot Q_i
//////////////////////////////////////////////////
void tobary(double *hPos, double *hVel, double *hPoshel, double *hVelhel, 
        int numBodies)
{
   double mtot = 0.0; // total mass
   double sumx[3] = {0.0,0.0,0.0}; // for center of mass 
   double sumv[3] = {0.0,0.0,0.0}; // total momentum
   int j; 
   int j0=4*0;
   double mi;
   if (hPos[3] == 0.0) printf("tobary: m0==0\n");
   for(int i=0;i<numBodies;i++){ 
      j = 4*i;
      mi=hPos[j+3];
      mtot += mi; 
      for (int k=0;k<3;k++){
         sumx[k] += mi*hPos[j+k]; sumv[k] += mi*hVel[j+k]; 
      }
      hPoshel[j+3] = mi; //remember to transfer masses
// note: mass in 0 position is m0 right now not mtot
   }
   for (int k=0;k<3;k++){  // zero  coord
      hVelhel[j0+k] = sumv[k]/mtot;  // Center of momemtum 
      hPoshel[j0+k] = sumx[k]/mtot;  // Center of mass
   }
   for(int i=1;i<numBodies;i++){
      j = 4*i;
      for (int k=0;k<3;k++){
        hVelhel[j+k] = hVel[j+k] - sumv[k]/mtot;  //eqn 30
        hPoshel[j+k] = hPos[j+k] - hPos[j0+k]; // eqn 27
      }
   }

}

//////////////////////////////////////////////////
// reverse transform
// equation 29 from Duncan et al. 98
//  hVelhel, hPoshel are inputs 
//  hPos, hVel are outputs 
//////////////////////////////////////////////////
void frombary(double *hPos, double *hVel, double *hPoshel, double *hVelhel, 
        int numBodies)
{
   double sumx[3] = {0.0,0.0,0.0}; // for center of mass
   double momv[3] = {0.0,0.0,0.0}; // total momentum
   double mtot=0.0; // total mass
   int j; 
   int j0=4*0;
   double m0 = hPoshel[j0+3]; // consistent with previous
   if (m0 == 0.0) printf("frombary: m0==0\n");
   mtot = m0; // to start 
   for(int i=1;i<numBodies;i++){
      j=4*i;
      double mi=hPoshel[j+3];
      mtot += mi; 
      for (int k=0;k<3;k++){ 
         sumx[k] += mi*hPoshel[j+k]; momv[k] += mi*hVelhel[j+k]; 
      }
      hPos[j+3] = hPoshel[j+3]; //remember to transfer masses
   }
   hPos[j0+3] = hPoshel[j0+3];  // transfer m0
   for (int k=0;k<3;k++){ 
       hPos[j0+k] = hPoshel[j0+k] - sumx[k]/mtot;
       hVel[j0+k] = hVelhel[j0+k] - momv[k]/m0;
   }
   for(int i=1;i<numBodies;i++){
      j=4*i;
      for (int k=0;k<3;k++){
        hPos[j+k] = hPoshel[j+k] + hPoshel[j0+k] - sumx[k]/mtot;
        hVel[j+k] = hVelhel[j+k] + hVelhel[j0+k];
      }
   }

}

// compute energy for cartesian coordinates on cpu
double energycart(double *hPos, double *hVel, int numBodies)
{
   double ke = 0.0;
   double pe = 0.0;
   for(int i=0;i<numBodies;i++){
      int ii=4*i;
      double v2=0.0;
      double mi = hPos[ii+3];
      for (int k=0;k<3;k++) v2+= hVel[ii+k]*hVel[ii+k];
      ke += 0.5*mi*v2;
   }
   for(int i=0;i<numBodies;i++){
      int ii=4*i;
      double mi = hPos[ii+3];
      for(int j=0;j<numBodies;j++){
        int jj=4*j;
        double mj = hPos[jj+3];
        double dr2=0.0;
        for (int k=0;k<3;k++) 
           dr2 += pow(hPos[ii+k] - hPos[jj+k],2.0);
        double dr = sqrt(dr2); 
        if (i!=j)  pe -= 0.5*mi*mj/dr;
      }
   }
   return pe+ke;
}
