#include <stdlib.h>
// #include <cutil.h>
#include <cuda.h>
#include <stdio.h>

// #include "doub.h"
#include "her_kernel.cu" // accelerations and jerks kernel
#include "nn_kernel.cu" // find nearest neighbor kernel

// forward declare
// give us a multiple of pthreads and larger or equal to nparticles
int gpusize(int nparticles, int pthreads){
   double cp = ceil((double)nparticles/(double)pthreads);
   return ((int)cp)*pthreads;
}

#define NTHREADS 128

//////////////////////////////////////////////////////////////////
// gravity routine for hermite integrator, calling GPU kernel
// this routine returns accelerations (forces)
// and their time derivatives (jerks) on all particles from all particles
// all particles reside in position,mass vector pos with velocities vel
// number of particles is nparticles
// Inputs:
//    h_Pos, h_Vel   array of particle positions and velocities 
//                   (includes masses as pow.w, on cpu)
//    eps2       softening^2
//    rc         not used right now           
//    nparticles number of particles,  numBodies is number sent to GPU
// Outputs:
//    h_acc, h_jrk   acceleration and jerks of all particles on cpu
// For the future:  
//      the device arrays should be allocated in the previous routine not here
//////////////////////////////////////////////////////
void gravitygpu(double4 *h_Pos, double4 *h_Vel, 
                double3 *h_acc, double3 *h_jrk,
                int nparticles, double eps2, double rc)
{
   cudaError_t err;
//   printf("beginning gravitygpu\n"); fflush(stdout);
   int p = NTHREADS; // hardset here
   dim3 pthreads(p,1,1);
   int numBodies =  gpusize(nparticles, p);
   // want numBodies to be multiple of p

   dim3 pgrid(numBodies/p,1,1);
   int sharedMemSize = 2*p*sizeof(double4);  // double4 s

   int h_memSize4 = sizeof(double4)*nparticles; // size on cpu
   int h_memSize3 = sizeof(double3)*nparticles;
   int d_memSize4 = sizeof(double4)*numBodies;  // size on gpu
   int d_memSize3 = sizeof(double3)*numBodies;
   double4 *d_Pos, *d_Vel;
   double3 *d_acc, *d_jrk;
   err = cudaMalloc((void**)&d_Pos,d_memSize4);
   if (err != cudaSuccess) printf("cudaMalloc fail d_Pos in gravitygpu\n");
   err = cudaMalloc((void**)&d_Vel,d_memSize4);
   if (err != cudaSuccess) printf("cudaMalloc fail d_Vel in gravitygpu\n");
   cudaMemset(d_Pos, 0, d_memSize4); // zero extra particles
   cudaMemset(d_Vel, 0, d_memSize4);
   err = cudaMalloc((void**)&d_acc,d_memSize3);
   if (err != cudaSuccess) printf("cudaMalloc fail d_acc in gravitygpu\n");
   err = cudaMalloc((void**)&d_jrk,d_memSize3);
   if (err != cudaSuccess) printf("cudaMalloc fail d_jrk in gravitygpu\n");
   cudaMemset(d_acc, 0, d_memSize3);
   cudaMemset(d_jrk, 0, d_memSize3);
   err = cudaMemcpy(d_Pos, h_Pos, h_memSize4,  
                              cudaMemcpyHostToDevice); //only copy nparticles
   if (err != cudaSuccess) printf("cudaMemcpy fail in gravitygpu\n");
   err = cudaMemcpy(d_Vel, h_Vel, h_memSize4, 
                              cudaMemcpyHostToDevice);
   if (err != cudaSuccess) printf("cudaMemcpy fail in gravitygpu\n");

   // run numBodies
   hergravity <<< pgrid, pthreads, sharedMemSize >>>
              (d_acc, d_jrk, d_Pos, d_Vel, eps2, numBodies);
   err = cudaThreadSynchronize();
   if (err != cudaSuccess) printf("hergravity kernel execution failed\n");

   err = cudaMemcpy(h_acc, d_acc, h_memSize3, 
                              cudaMemcpyDeviceToHost); // copy nparticles
   if (err != cudaSuccess) printf("cudaMemcpy fail in gravitygpu\n");
   err = cudaMemcpy(h_jrk, d_jrk, h_memSize3, 
                              cudaMemcpyDeviceToHost);
   if (err != cudaSuccess) printf("cudaMemcpy fail in gravitygpu\n");

// free memory on gpu
   err = cudaFree(d_Pos);
   if (err != cudaSuccess) printf("cudaFree fail in gravitygpu\n");
   err = cudaFree(d_Vel);
   if (err != cudaSuccess) printf("cudaFree fail in gravitygpu\n");
   err = cudaFree(d_acc);
   if (err != cudaSuccess) printf("cudaFree fail in gravitygpu\n");
   err = cudaFree(d_jrk);
   if (err != cudaSuccess) printf("cudaFree fail in gravitygpu\n");
//   printf("ending gravitygpu\n"); fflush(stdout);

}

/////////////////////////////////////////////////////
// return a vector of indices of nearest neighbors
// input:  the positions and the number of particles
// output: h_cindex, the list of incies
void findnngpu(double4 *h_Pos, int nparticles, int* h_cindex)
{
   cudaError_t err;
   printf("beginning findnngpu\n"); fflush(stdout);
   int p = NTHREADS; // hardset here
   dim3 pthreads(p,1,1);
   int numBodies =  gpusize(nparticles, p);
   // want numBodies to be multiple of p

   dim3 pgrid(numBodies/p,1,1);
   int sharedMemSize = p*sizeof(float4);  

   int h_memSize4 = sizeof(double4)*nparticles; // size on cpu
   int d_memSize4 = sizeof(double4)*numBodies;  // size on gpu
   double4 *d_Pos;
   err = cudaMalloc((void**)&d_Pos,d_memSize4);
   if (err != cudaSuccess) printf("cudaMalloc fail d_Pos in findnngpu\n");
   cudaMemset(d_Pos, 0, d_memSize4); // zero extra particles
   err = cudaMemcpy(d_Pos, h_Pos, h_memSize4,  
                              cudaMemcpyHostToDevice); //only copy nparticles
   if (err != cudaSuccess) printf("cudaMemcpy fail in findnngpu\n");
   int *d_cindex;
   int h_memSizei = sizeof(int)*nparticles; // size on cpu
   int d_memSizei = sizeof(int)*numBodies;  // size on gpu
   err = cudaMalloc((void**)&d_cindex,d_memSizei);
   if (err != cudaSuccess) printf("cudaMalloc fail d_cindex in findnngpu\n");
   cudaMemset(d_cindex, 0, d_memSizei); // zero 

   // run the kernel
   findnearest <<< pgrid, pthreads, sharedMemSize >>>
              (d_Pos, numBodies, d_cindex);
   err = cudaThreadSynchronize();
   if (err != cudaSuccess) printf("findnearest kernel execution failed\n");
   err = cudaMemcpy(h_cindex, d_cindex, h_memSizei,  
                              cudaMemcpyDeviceToHost); //only copy nparticles
   if (err != cudaSuccess) printf("cudaMemcpy fail in findnngpu\n");

// free memory on gpu
   err = cudaFree(d_Pos);
   if (err != cudaSuccess) printf("cudaFree fail in gravitygpu\n");
   err = cudaFree(d_cindex);
   if (err != cudaSuccess) printf("cudaFree fail in gravitygpu\n");

   printf("ending findnngpu\n"); fflush(stdout);

}
