

// all force pair computations, accelerations and jerks,  
// for hermite integrator
// needed to simulate gravity() routine but on gpu
 

// #include "doub_all.h"
 
#ifndef _HER_KERNEL_H_
#define _HER_KERNEL_H_

#include <host_defines.h>

#define LOOP_UNROLL 4

/*
typedef struct __align__(16) {
   double ax,ay,az,jx,jy,jz;
} accjrk;
*/

struct accjrk  { double ax,ay,az,jx,jy,jz; } ;


///////////////////////////////////////////////////////////
// soft2 is softening squared
// add acceleration and jrk to accjrk ai on particle bi from particle bj
// mass for bj is stored in bj.w
// bi,bj are positions of particle b i,j
// bvi,bvj are velocites of particle b i,j
__device__ accjrk  
bodyBodyInteraction(accjrk ai, double4 bi, double4 bj, 
                               double4 bvi, double4 bvj, 
                               double soft2) 
{
    double3 dr,dv;

    dr.x = bj.x - bi.x;   dv.x = bvj.x - bvi.x;
    dr.y = bj.y - bi.y;   dv.y = bvj.y - bvi.y;
    dr.z = bj.z - bi.z;   dv.z = bvj.z - bvi.z;
    double distSqr = dr.x*dr.x + dr.y*dr.y + dr.z*dr.z;
    distSqr += soft2;
    double invDist = 1.0/sqrt(distSqr);
    double invDistCube =  invDist * invDist * invDist;
    double s = bj.w * invDistCube;
    double rdotv = dr.x*dv.x + dr.y*dv.y + dr.z*dv.z;
    double threerdotvdr2 = 3.0*rdotv/distSqr;

    ai.ax += dr.x *s;
    ai.ay += dr.y *s;
    ai.az += dr.z *s;
    ai.jx += s*(dv.x - threerdotvdr2*dr.x); 
    ai.jy += s*(dv.y - threerdotvdr2*dr.y); 
    ai.jz += s*(dv.z - threerdotvdr2*dr.z); 

    return ai;
}


/////////////////////////////////////////////////////////
// add forces/jrks from every particle looping through the blockdim
// This is a "tile_calculation" function 
/////////////////////////////////////////////////////////
__device__ accjrk gravitation(double4 myPos, double4 myVel, 
           accjrk accel, double soft2)
{
    extern __shared__ double4 sharedPos[];
    double4* sharedVel;
    sharedVel = &sharedPos[blockDim.x];
    int i=0;

    for (i = 0; i < blockDim.x; ) 
    {
//        accel = bodyBodyInteraction(accel,myPos,sharedPos[i++],soft2); 
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i], myVel, 
                         sharedVel[i], soft2); i++;
#if LOOP_UNROLL > 1
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i], myVel, 
                         sharedVel[i], soft2); i++;
#endif
#if LOOP_UNROLL > 2  
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i], myVel, 
                         sharedVel[i], soft2); i++;
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i], myVel, 
                         sharedVel[i], soft2); i++;
#endif
#if LOOP_UNROLL > 4
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i], myVel, 
                         sharedVel[i], soft2); i++;
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i], myVel, 
                         sharedVel[i], soft2); i++;
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i], myVel, 
                         sharedVel[i], soft2); i++;
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i], myVel, 
                         sharedVel[i], soft2); i++;
#endif
    }

    return accel;
}

// WRAP is used to force each block to start working on a different 
// chunk (and wrap around back to the beginning of the array) so that
// not all multiprocessors try to read the same memory locations at 
// once.
#define WRAP(x,m) (((x)<m)?(x):(x-m))  // Mod without divide, works on values from 0 up to 2m

////////////////////////////////////////////////////////////////
// Nylands from Cuda 1.1 SDK
// this does not depend on blockIdx
////////////////////////////////////////////////////////////////
__device__ accjrk 
computeBodyAccel(double4 bodyPos, double4 bodyVel, 
              double4* positions, double4* velocities,
              int numBodies, double soft2)
{
    extern __shared__ double4 sharedPos[];
    double4* sharedVel;
    sharedVel = &sharedPos[blockDim.x];

    accjrk acc = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0}; // zero initialize
    
    int p = blockDim.x;
    int n = numBodies;  // note numBodies could be num massive here
                        // and shorter than the total as long as mod p
    int i,tile;
   // number of tiles is N/p = blockdim.x which is why wrap works

    __syncthreads();
// load up mem for p particles then compute forces 
// the memory is transfered over and over again 
// but arranged so there is no interference
// between memory transfers

    for (i = 0, tile = 0; i < n; i += p, tile++)  // looping with step of tile (p) 
    { // load up shared memory 
        int idx = WRAP(blockIdx.x+tile, gridDim.x) * blockDim.x + threadIdx.x;
        sharedPos[threadIdx.x] = positions[idx];
        sharedVel[threadIdx.x] = velocities[idx];
       
        __syncthreads();
        // This is the "tile_calculation" function 
          acc = gravitation(bodyPos, bodyVel, acc, soft2);
        __syncthreads();
    }

    return acc;
}

////////////////////////////////////////////////////////////////
// return an array of jerks and accelerations summed over all particles
////////////////////////////////////////////////////////////////
__global__ void hergravity(double3* acc, double3* jrk, 
                double4* d_Pos, double4* d_Vel, double soft2,
                int numBodies)
{
    extern __shared__ double4 sharedPos[];
    int index = blockIdx.x*blockDim.x + threadIdx.x;
    double4 vel = d_Vel[index];
    double4 pos = d_Pos[index];   

    accjrk acc6 = computeBodyAccel(pos, vel, d_Pos, d_Vel, 
            numBodies, soft2);
    acc[index].x = acc6.ax;
    acc[index].y = acc6.ay;
    acc[index].z = acc6.az;
    jrk[index].x = acc6.jx;
    jrk[index].y = acc6.jy;
    jrk[index].z = acc6.jz;

    __syncthreads();
}


#endif // #ifndef _HER_KERNEL_H_

