
/* 
take a keplerian step for all particles on the gpu
following recipe by Prussing and Conway
*/

#ifndef _KEP_KERNEL_H_
#define _KEP_KERNEL_H_


// #include "doub_all.h"

// forward declares
__device__ double solvex(double r0dotv0, double alpha,
                double M1, double r0, double dt);
__device__ double C_prussing(double y);
__device__ double S_prussing(double y);
__device__ void kepstep(double4 pos, double4 vel,
                double4* npos, double4* nvel,
                double deltaTime, double GM);


//////////////////////////////////////////////////////
// advance all particles
// calling a kepstep
// new positions and vels into newPos, newVel
// old ones not changed
// number of bodies passed but not used
// does not matter what first body is, will not move it
//////////////////////////////////////////////////////
__global__ void kepstepBodies(double4* newPos, double4* newVel,
                double4* oldPos, double4* oldVel,
                double deltaTime, int numBodies, double GM)
{
   int index = blockIdx.x*blockDim.x + threadIdx.x;
   double4 pos = oldPos[index];
   double4 vel = oldVel[index];
   double4 npos = {0.0, 0.0, 0.0, 0.0};
   double4 nvel = {0.0, 0.0, 0.0, 0.0};

   kepstep(pos, vel, &npos, &nvel, deltaTime, GM);

 // store new position and velocity
    if (index > 0){  // prevent infinity if first body is at origin
      newPos[index] = npos; 
      newVel[index] = nvel;
    }
    else {
      newPos[index] = pos; // don't allow Sun to move
      newVel[index] = vel;
    }
    __syncthreads();
}


///////////////////////////////////////////////////////////////
// advance a particle using f,g functions and universal variables
// for differental kepler's equation
//  has an error catch for r0=0 so can be run with central body
// Based on equations by Prussing, J. E. and Conway, B. A. 
// Orbital Mechanics 1993, Oxford University Press, NY,NY  chap 2 
// npos,nvel are new positions and velocity
// pos, vel are old ones
///////////////////////////////////////////////////////////////
#define MINR 1.0e-5 // minimum radius
#define MINDENOM 1e-8  // mininum denominator
__device__ void kepstep(double4 pos, double4 vel,
                double4* npos, double4* nvel,
                double deltaTime, double GM)
{

   double r0 = sqrt(pos.x * pos.x + pos.y * pos.y + pos.z * pos.z + MINR*MINR); // current radius
   double v2 = (vel.x * vel.x + vel.y * vel.y + vel.z * vel.z);  // current velocity
   double r0dotv0 = (pos.x*vel.x + pos.y*vel.y + pos.z*vel.z);
   double alpha = (2.0/r0 - v2/GM);  // inverse of semi-major eqn 2.134 MD
// here alpha=1/a and can be negative
   double x_p = solvex(r0dotv0, alpha, GM, r0, deltaTime); // solve universal kepler eqn

   double smu = sqrt(GM);
   double foo = 1.0 - r0*alpha;
   double sig0 = r0dotv0/smu;
   double x2 = x_p*x_p;
   double x3 = x2*x_p;
   double alx2 = alpha*x2;
   double Cp = C_prussing(alx2);
   double Sp = S_prussing(alx2);
   double r = sig0*x_p*(1.0 - alx2*Sp)  + foo*x2*Cp + r0; // eqn 2.42  PC
   if (r < MINR) r=MINR;
// if dt == 0 then f=dgdt=1 and g=dfdt=0
// f,g functions equation 2.38a  PC
   double f_p= 1.0 - (x2/r0)*Cp;
   double g_p= deltaTime - (x3/smu)*Sp;
// dfdt,dgdt function equation 2.38b PC
   double dfdt;
   double dgdt = 1.0 - (x2/r)*Cp;
   if (fabs(g_p) > MINDENOM)
      // conservation of angular momentum means that f dfdt - g dfdt =1
      dfdt = (f_p*dgdt - 1.0)/g_p;
   else
      // dfdt,dgdt function equation 2.38b PC
      dfdt = x_p*smu/(r*r0)*(alx2*Sp - 1.0);
  
//   double4 npos; // new ones
//   double4 nvel;
   npos->x = f_p*pos.x + g_p*vel.x; // eqn 2.65 M+D
   npos->y = f_p*pos.y + g_p*vel.y;
   npos->z = f_p*pos.z + g_p*vel.z;
   npos->w = pos.w; // otherwise mass lost
   nvel->x = dfdt*pos.x + dgdt*vel.x; //eqn 2.70 M+D
   nvel->y = dfdt*pos.y + dgdt*vel.y;
   nvel->z = dfdt*pos.z + dgdt*vel.z;

}

////////////////////////////////////////////////////////////////
// solving differential Kepler's equation
// in universal variable x
// using Laguerre method as outlined by Prusing+C eqn 2.43
////////////////////////////////////////////////////////////////
#define N_LAG 5.0 // integer n, for recommended Laguerre method
#define SIGN(a) ((a) < 0 ? -1 : 1)

__device__ double solvex(double r0dotv0, double alpha,
                double M1, double r0, double dt)
{
   double smu = sqrt(M1);
   double foo = 1.0 - r0*alpha;
   double sig0 = r0dotv0/smu;
   double x = M1*dt*dt/r0; // initial guess could be improved 
// better initial guess depends on rperi which would have to be passed

   double u=1.0;
   for(int i=0;i<7;i++){  // 7 iterations is probably overkill
			// as it always converges faster than this
     double x2,x3,alx2,Cp,Sp,F,dF,ddF,z;
     x2 = x*x;
     x3 = x2*x;
     alx2 = alpha*x2;
     Cp = C_prussing(alx2);
     Sp = S_prussing(alx2);
     F = sig0*x2*Cp + foo*x3*Sp + r0*x - smu*dt; // eqn 2.41 PC
     dF = sig0*x*(1.0 - alx2*Sp)  + foo*x2*Cp + r0; // eqn 2.42 PC
     ddF = sig0*(1.0-alx2*Cp) + foo*x*(1.0 - alx2*Sp);
     z = fabs((N_LAG - 1.0)*((N_LAG - 1.0)*dF*dF - N_LAG*F*ddF));
     z = sqrt(z);
     double denom = (dF + SIGN(dF)*z); 
     if (denom ==0.0) denom = MINDENOM;
     u = N_LAG*F/denom; // equation 2.43 PC
     x -= u;
   }
//   if (isnan(x)) printf("solvex: is nan\n");
   return x;
}

// functions needed
__device__ double C_prussing(double y) // equation 2.40a Prussing + Conway
{
  if (fabs(y)<1e-4) return 1.0/2.0*(1.0 - y/12.0*(1.0 - y/30.0*(1.0 - y/56.0)));
  double u = sqrt(fabs(y));
  if (y>0.0) return (1.0- cos(u))/ y;
  else       return (cosh(u)-1.0)/-y;
}

__device__ double S_prussing(double y) // equation 2.40b Prussing +Conway
{
  if (fabs(y)<1e-4) return 1.0/6.0*(1.0 - y/20.0*(1.0 - y/42.0*(1.0 - y/72.0)));
  double u = sqrt(fabs(y));
  double u3 = u*u*u;
  if (y>0.0) return (u -  sin(u))/u3;
  else       return (sinh(u) - u)/u3;
}

                               
#endif

