#include <stdlib.h>
// #include <cutil.h>
#include <cuda.h>
#include <stdio.h>
#include <math.h>

//#include "doub.h"

#include "sweep1_kernel.cu" // sweep1,sweep2,sweep3 kernels

// get a list of pairs undergoing close approaches during timestep
// mkpairlist

// calls this:
void scanpass(uint *d_odata, uint* d_idata, uint numElements);

extern "C"
{



// mem requirements on GPU
//  d_npass, d_npass_scan    numBodies*sizeof(int)  x2 because 2 of them
//  d_pairmark, d_pairscan    npairs1*sizeof(uint) x2
//  d_pairlist sizeof(long long int)
//  d_newpairlist   npairs2*sizeof(long long int)
// npairs1 is from first sweep
// npairs2 is final number , probably called newnpairs

#define NTHREADS 128 // hardset here

//////////////////////////////////////////////////////////////////
// Search for particles undergoing close approaches in time deltaTime
// Inputs  on device:
//    m_dPos, m_dVel at t=0   particle positions and velocities
//    m_dPos, m_dVel at t=deltaTime
// Inputs On Host
//    numBodies     number of particles
//    numPlanets    only looking for interactions that include these
//                  should be an integer multiple of pthreads
//    GM            m0
//    deltaTime     timestep 
// Outputs
//    h_newpairlist    with pairids in it
//                  allocated on cpu, allocated here, remember to free later 
// Returning:  npairs in pairlist
//
// Kernels used:   sweep1, sweep2, sweep3p
//        also calls scan from cudpp
// currently all particles called 
// particles from offset to numPlanets are considered for interactions
// could expand this so can be called in pieces and
//        then pairlists merged at the end
//////////////////////////
int mkpairlist(double* d_Pos0, double* d_Vel0, double* d_Pos1, double *d_Vel1,
        int numBodies, int numPlanets, long long int** h_newpairlist, 
        double GM,double m0, double deltaTime, double hillfac)
{
   cudaError_t err;
   int offset=0; // currently only search top 
   int p = NTHREADS; // hardset here
   if (numBodies%p !=0) {
      printf("numBodies %d not an integer multiple of pthreads in mkpairlist\n"
          ,numBodies);
   }
   dim3 pthreads(p,1,1);
   dim3 pgrid(numBodies/p,1,1);
   int sharedMemSize = p*sizeof(double4);  // double4s
   uint* d_npass, *d_npass_scan; 
   int memSize = sizeof(uint)*numBodies;
   err = cudaMalloc((void**)&d_npass,memSize);
   if (err != cudaSuccess) printf("cudaMalloc fail d_npass in mkpairlist \n");
   err = cudaMalloc((void**)&d_npass_scan,memSize);
   if (err != cudaSuccess) printf("cudaMalloc fail d_npass_scan in mkpairlist \n");
   uint *h_npass;
   h_npass = new uint[numBodies];
   memset(h_npass, 0, numBodies*sizeof(uint));
// zero the gpu arrays because we might run the routines for fewer particles
// than numBodies
   err = cudaMemcpy(d_npass,      h_npass, numBodies*sizeof(uint), 
                              cudaMemcpyHostToDevice);
   if (err != cudaSuccess) printf("cudaMemcpy fail d_npass in mkpairlist \n");
   err = cudaMemcpy(d_npass_scan, h_npass, numBodies*sizeof(uint), 
                              cudaMemcpyHostToDevice);
   if (err != cudaSuccess) printf("cudaMemcpy fail d_npass_scan in mkpairlist \n");
// find out how many interactors per particle
   sweep1 <<< pgrid, pthreads, sharedMemSize >>>
        (d_npass, (double4*)d_Pos0, (double4*)d_Vel0, 
         deltaTime, numBodies,numPlanets, offset, GM,m0, hillfac);
   err = cudaThreadSynchronize();
   if (err != cudaSuccess) printf("sweep1 kernel execution failed\n");


// for debugging
// err = cudaMemcpy(h_npass, d_npass, numBodies*sizeof(uint), 
//                            cudaMemcpyDeviceToHost);
   
// scan d_pass --- d_pass has number of interacting per particle
   scanpass(d_npass_scan,d_npass,numBodies);
// get the scanned array
   err = cudaMemcpy(h_npass, d_npass_scan, numBodies*sizeof(uint), 
                              cudaMemcpyDeviceToHost);
   if (err != cudaSuccess) printf("cudaMemcpy fail h_npass in mkpairlist \n");
// printf("mkpl: after first sweep and scan\n");
// figure out how many pairs from first sweep
   uint npairs = h_npass[numBodies-1];
   if (npairs > 1000) // warning just in case
     printf("mkpl: npairs first sweep total=%d\n",npairs);
   
   delete [] h_npass; // cleanup
   err = cudaFree(d_npass);
   if (err != cudaSuccess) printf("cudaFree fail in mkpairlist\n");

// setup for second sweep and making the pair list
   long long int *d_pairlist; 
   memSize = sizeof(long long int)*(npairs+1);
   err = cudaMalloc((void**)&d_pairlist,memSize);
   if (err != cudaSuccess) printf("cudaMalloc fail &d_pairlist in mkpairlist \n");
// fill up the pair list
   sweep2 <<< pgrid, pthreads, sharedMemSize >>>
        (d_pairlist, (double4*)d_Pos0, (double4*)d_Vel0, 
         deltaTime, d_npass_scan,  numBodies, 
         numPlanets, offset, GM,m0,hillfac);
   
   err = cudaThreadSynchronize();
   if (err != cudaSuccess) printf("sweep2 kernel execution failed\n");
//   printf("mkpl: after second sweep, filled pairlist \n");
   
// free mem no longer needed 
   err = cudaFree(d_npass_scan);
   if (err != cudaSuccess) printf("cudaFree fail in mkpairlist\n");

// setup to do a finer sweep on the pairlist
   memSize = sizeof(uint)*(npairs+1);
   uint *d_pairmark,*d_pairscan;
   err = cudaMalloc((void**)&d_pairmark,memSize);
   if (err != cudaSuccess) printf("cudaMalloc fail &d_pairmark in mkpairlist \n");
   err = cudaMalloc((void**)&d_pairscan,memSize);
   if (err != cudaSuccess) printf("cudaMalloc fail &d_pairscan in mkpairlist \n");
   uint *h_pairscan;
   h_pairscan = new uint[npairs+1];

   double cp = ceil((double)npairs/(double)p);
//   printf("cp=%d %.2f\n",(int)cp,cp);
   dim3 qgrid((int)cp,1,1);
// do a finer sweep removing more pairs that don't approach
// using Chambers' routine
   sweep3p <<< qgrid, pthreads,0  >>>
        ((double4*)d_Pos0, (double4*)d_Vel0, 
         (double4*)d_Pos1, (double4*)d_Vel1, 
         deltaTime, numBodies, npairs, d_pairlist, 
         d_pairmark, GM, hillfac);
   err = cudaThreadSynchronize();
   if (err != cudaSuccess) printf("sweep3 kernel execution failed\n");
//   printf("mkpl: after third sweep and pairmark \n");

// count how many pairs left after third sweep
   scanpass(d_pairscan,d_pairmark,npairs+1);
   err = cudaMemcpy(h_pairscan, d_pairscan, (npairs+1)*sizeof(uint), 
                              cudaMemcpyDeviceToHost);
   if (err != cudaSuccess) printf("cudaMemcpy fail h_pairscan in mkpairlist \n");
   int newnpairs = h_pairscan[npairs];
//   printf("mkpl: new npairs=%d\n",newnpairs);
   delete [] h_pairscan;
   long long int *d_newpairlist;
   // memSize = sizeof(uint)*(newnpairs+1);
   memSize = sizeof(long long int)*(npairs+1);
   err = cudaMalloc((void**)&d_newpairlist,memSize);
   if (err != cudaSuccess) printf("cudaMalloc fail &d_newpairlist in mkpairlist \n");
// repack the array to a smaller one
   repack <<< qgrid, pthreads,0  >>>
       (d_newpairlist, d_pairlist, d_pairscan, npairs, newnpairs);
   err = cudaThreadSynchronize();
   if (err != cudaSuccess) printf("repack kernel execution failed\n");
//   printf("mkpl: repack of pairlist done\n");


// free the stuff for the old pairlist
   err = cudaFree(d_pairscan);
   if (err != cudaSuccess) printf("cudaFree fail in mkpairlist\n");
   err = cudaFree(d_pairmark);
   if (err != cudaSuccess) printf("cudaFree fail in mkpairlist\n");
   err = cudaFree(d_pairlist);
   if (err != cudaSuccess) printf("cudaFree fail in mkpairlist\n");

// written now with malloc instead of new/delete
   *h_newpairlist = (long long int *)malloc((newnpairs+1)*sizeof(long long int));

// get the new pairlist onto host
   err = cudaMemcpy(*h_newpairlist, d_newpairlist, 
          (newnpairs+1)*sizeof(long long int),cudaMemcpyDeviceToHost);
   if (err != cudaSuccess) printf("cudaMemcpy fail h_newpairlist in mkpairlist \n");
// clean up
   err = cudaFree(d_newpairlist);
   if (err != cudaSuccess) printf("cudaFree fail in mkpairlist\n");

   return newnpairs;
}

}
