// useful structure to store information about the simulation

#ifndef _KEPCART_
#define _KEPCART_

typedef struct {
  double x, y, z, xd, yd, zd;
} PhaseState;

typedef struct {
  double a, e, i, longnode, argperi, meananom;
} OrbitalElements;

#endif 

#include "vector_types.h"

#define npmax 10 // maximum number of planets!
struct SIMinfo
{
  char outroot[30]; // root output file name
  int numBodies;
  int numBodiesMassive;
  int device;       // which GPU
  double deltaTime; // timestep interval
  int niter;       // total number of iterations intended
  int ndt;         // number of timesteps per interation
  double damping;   // damping coefficient if any
  double soft2_int; // softening coefficient soft2 in interaction step
  double soft2_her; // softening coefficient soft2 in hermite integrator
  double hillfac_int; // hillfactor used in interaction term and hermite int
  double hillfac_col; // hillfactor used in encounter detection
  double amin;     // for particle maker semi-major axis range
  double amax;     // semi-major axis range
  double gamma;    // powerlaw distribution p(a) ~ a^gamma
  double ibar;     // sqrt(<i^2>)  inclination dispersion
  double m;        // mass of non-central particles
  int    nplanets; // number of initial planets
  double mplanets[npmax];  // mass of planets (not star)
  OrbitalElements orbel[npmax];  // initial orbital elements w.r.t star
  double time;    // current simulation time since start
};

// HD5part reading and writing subroutines
// timeStep =0 is first one
// routines in iostuff.cpp

#ifdef H5PART
  void write_h5part(char *filename, double *Pos, double *Vel, SIMinfo siminfo, int timeStep);
  void info_h5part(char *filename, SIMinfo *siminfo);
  void read_h5part(char *filename, double *Pos, double *Vel, int timeStep);
#endif

  void write_asc(char *filename, double *Pos, double *Vel, SIMinfo siminfo, int timeStep);
  void info_asc(char *filename, SIMinfo *siminfo);
  void read_asc(char *filename, double *Pos, double *Vel, int timeStep);


// from sread.cpp
void sread(char *parmfile,  SIMinfo *siminfo);

// from coord.cpp
void correctcom(double *hPos, double *hVel, int numBodies);
void tobary(double *hPos, double *hVel, double *hPoshel, double *hVelhel,
        int numBodies);
void frombary(double *hPos, double *hVel, double *hPoshel, double *hVelhel,
        int numBodies);
// from testcode.cpp
double doenn(double* pos, double *vel,  int numBodies, double soft2);

// from mklists.cpp
int hermlists(int* indexlist, int numBodies, 
          long long int* pairlist, int npairs);
int nhermlists(int* indexlist, int* startlist, int* nencounter,
              int numBodies, long long int* pairlist, int npairs);
void printenclist(int* indexlist, int np);
void printallenclist(int* indexlist, int* startlist, int nlist);
void printtocoll(char* pstring);

// from hermitep.cpp
void callHermite(int *indexlist, double4* hPos, double4 *hVel, int nindexes,
     double dtau, double eps2, double hillfac, double m0, 
     int dmovesun, double time, int numPlanets);
void callIntCor(int *indexlist, double4* hPos, double4* hVel, int nindexes,
     double dtau, double eps2, double hillfac);
double energy(double4 *h_Pos, double4 *h_Vel, int nparticles, double soft2, 
            double rc);
double Edrift(double4 *hPos, double4 *hVel, int nparticles);
double callHill(int *indexlist, int nindexes, double4 *h_Pos, double m0);
void hermiteBodies(double4 *pos, double4 *vel,
       int nparticles, double tstep, double eps2, double rc, int dmovesun, 
       double time, int* indexlist, int numPlanets);
double largesthill(double4 *h_Pos, int nparticles);

// extern "C"  //  from coll.cpp  or colltest.cpp

void integrateKepCol(double* aPos, double* aVel,
                double* bPos, double* bVel,
                double deltaTime, double damping,
                double soft2_int, double soft2_herm,
                double hillfac_int, double hillfac_col,
                int numBodies, int numBodiesMassive,  int numPlanets,
                int p, double m0, double time);




extern "C" 
{
// from mkpairlist.cu
int mkpairlist(double* d_Pos0, double* d_Vel0, double* d_Pos1, double *d_Vel1,
        int numBodies, int numPlanets, long long int** h_pairlist,
        double GM, double m0, double deltaTime, double hillfac);

 //  these routines from bodysystemcuda.cu

    void checkCUDA(int device);
    void setDeviceSoftening(double softening);
    int allocateNBodyArrays(double* vel[2], int numBodies);
    void deleteNBodyArrays(double* vel[2]);
    void copyArrayFromDevice(double* host, const double* device, int numBodies);
    void copyArrayToDevice(double* device, const double* host, int numBodies);

    void integrateNbodySystem(double* newPos, double* newVel,
                              double* oldPos, double* oldVel,
                              double deltaTime, double damping, double soft,
                              int numBodies, int p);
    double energysum(double* Pos,  double* Vel,
                     double soft2, int numBodies, int p, double m0);
    void kepstepNbodySystem(double* newPos, double* newVel,
                     double* oldPos, double* oldVel,
                     double deltaTime,
                     int numBodies, int p, double m0);
    void sumMomenta(double* Pos, double* Vel, int p, int numBodies, double *pxpypz);
    void kickstepNbodySystem(double* newPos, double* newVel,
                     double* oldPos, double* oldVel,
                     double deltaTime, double damping,  double soft2,
                     int numBodies, int numBodiesMassive, int p, double m0);
    void driftstepNbodySystem(double* newPos, double* newVel,
                     double* oldPos, double* oldVel,
                     double deltaTime,
                     int numBodies, int p);
    void integrateKepNbodySystem(double* aPos, double* aVel,   
                     double* bPos, double* bVel,
                     double deltaTime, double damping,  double soft2,
                     int numBodies, int numBodiesMassive, int p, double m0);
    void callchangem0(double *aPos, double m0);

}


