

/* all force pair computations 
   Device code. force computations are modelled from 
   based on GEMS3 implimentation by L. Nyland
   and trimmed version of nbody in SDK
   The citation is:
     Fast N-body simulation with CUDA,
     Nyland, L.,  Harris,  M., \& and Prins, J.  2008, Chap 31
     in GPUGems3, edited by Hubert Nguyen,
     2008, Addison-Wesley,  Upper Saddle River, NJ, page 677

   The integrator has been modified to be leapfrog.  
          This is not called in symplectic integrator
   I have also added a kick step routine useful for symplectic integrators.
 */

// #include "doub_all.h"
 
#ifndef _NBODY_KERNEL_H_
#define _NBODY_KERNEL_H_

#include <host_defines.h>

#define LOOP_UNROLL 4
// garbage if larger than 2

//__constant__ double softeningSquared;  // currently not used



///////////////////////////////////////////////////////////
// soft2 is softening squared
// add acceleration to ai on particle bi from particle bj
// mass for bj is stored in bj.w
__device__ double3 
bodyBodyInteraction(double3 ai, double4 bi, double4 bj, double soft2) 
{
    double3 r;

    // r_ij  [3 FLOPS]  I switched sign here from that in SDK
    r.x = bj.x - bi.x;
    r.y = bj.y - bi.y;
    r.z = bj.z - bi.z;

    // distSqr = dot(r_ij, r_ij) + EPS^2  [6 FLOPS]
    double distSqr = r.x * r.x + r.y * r.y + r.z * r.z;
    distSqr += soft2;

    double invDist = 1.0/sqrt(distSqr);
    double invDistCube =  invDist * invDist * invDist;

    //  m_j * invDistCube [1 FLOP]
    double s = bj.w * invDistCube;

    // a_i =  a_i + s * r_ij [6 FLOPS]
    ai.x += r.x * s;
    ai.y += r.y * s;
    ai.z += r.z * s;

    return ai;
}

/////////////////////////////////////////////////////////
// add forces from every particle looping through the blockdim
// This is the "tile_calculation" function from the GPUG3 article.
// howerver SDK version had arguments flipped
/////////////////////////////////////////////////////////
__device__ double3 gravitation(double4 myPos, double3 accel, double soft2)
{
    extern __shared__ double4 sharedPos[];
    int i=0;

    // Here we unroll the loop
    // what if blockDim.x is not divisible by LOOP_UNROLL? 
    // typically blockDim.x is 256 threads and no problem
    // for reasons I can't figure out this routine crashes when looproll >2

    for (i = 0; i < blockDim.x; ) 
    {
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i],soft2);  i++;
               // I think flipped in SDK 
#if LOOP_UNROLL > 1
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i],soft2);  i++;
#endif
#if LOOP_UNROLL > 2  
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i],soft2);  i++;
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i],soft2);  i++;
#endif
#if LOOP_UNROLL > 4
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i],soft2);  i++;
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i],soft2);  i++;
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i],soft2);  i++;
        accel = bodyBodyInteraction(accel,myPos,sharedPos[i],soft2);  i++;
#endif
    }

    return accel;
}

// WRAP is used to force each block to start working on a different 
// chunk (and wrap around back to the beginning of the array) so that
// not all multiprocessors try to read the same memory locations at 
// once.
#define WRAP(x,m) (((x)<m)?(x):(x-m))  // Mod without divide, works on values from 0 up to 2m

////////////////////////////////////////////////////////////////
// Nylands from Cuda 1.1 SDK
// this does not depend on blockIdx
// There will be an error if numBodies is not an integer multiple
// as previous loads into shared mem will still be used.
////////////////////////////////////////////////////////////////
__device__ double3
computeBodyAccel(double4 bodyPos, double4* positions, int numBodies, double soft2)
{
    extern __shared__ double4 sharedPos[]; 

    double3 acc = {0.0, 0.0, 0.0}; // zero initialize 
    
    int p = blockDim.x;
    int n = numBodies;  // note numBodies could be num massive here
                        // and shorter than the total as long as mod p =0
    int i,tile;
   // number of tiles is N/p = blockdim.x which is why wrap works

    __syncthreads();
// load up mem for p particles then compute forces 
// the memory is transfered over and over again 
// but arranged so there is no interference
// between memory transfers
 
    for (i = 0, tile = 0; i < n; i += p, tile++)  // looping with step of tile (p)  
    { // load up shared memory  
        int idx = tile*blockDim.x + threadIdx.x; 
        sharedPos[threadIdx.x] = positions[idx]; 
//      sharedPos[threadIdx.x] =  
//          positions[WRAP(blockIdx.x+tile, gridDim.x) * blockDim.x + threadIdx.x]; 
// no difference between these 2 seen 
        
        __syncthreads(); 
        // This is the "tile_calculation" function from the GPUG3 article.
          acc = gravitation(bodyPos, acc, soft2); 
        __syncthreads(); 
    } 
 
    return acc; 
} 
 
////////////////////////////////////////////////////////////////
// does one timestep updating both positions and velocities with leapfrog
// not used right now in symplectic integrator
////////////////////////////////////////////////////////////////
__global__ void integrateBodies(double4* newPos, double4* newVel, 
                double4* oldPos, double4* oldVel,
                double deltaTime, double damping, double soft2,
                int numBodies)
{
    extern __shared__ double4 sharedPos[];
    int index = blockIdx.x*blockDim.x + threadIdx.x;
    double4 vel = oldVel[index];
    double4 pos = oldPos[index];   

    double3 accel = computeBodyAccel(pos, oldPos, numBodies, soft2);

    // new velocity = old velocity + acceleration * deltaTime
    // here force == acceleration
       
   // first update velocity 
    vel.x += accel.x * deltaTime;
    vel.y += accel.y * deltaTime;
    vel.z += accel.z * deltaTime; 
// could add here a central force law or halo term
// double r = sqrt(pos.x*pos.x + pos.y*pos.y + pos.z*pos.z);
// if adding kep term use bodyBodyInteraction to calculate accel

// should be 1.0 if no damping
//    vel.x *= damping;  vel.y *= damping;  vel.z *= damping;
        
   // new position = old position + newvelocity * deltaTime
   // update positions based on the new updated velocity 
   // so is essentially a leap frog 
   __syncthreads(); // needed so positions of some particles 
                    // don't change before
                    // all velocities updated.

    pos.x += vel.x * deltaTime;
    pos.y += vel.y * deltaTime;
    pos.z += vel.z * deltaTime;

// store new position and velocity
    newPos[index] = pos;
    newVel[index] = vel;
    __syncthreads();

}

///////////////////////////////////////////////////////////
// does velocity kicks only, for interaction step, positions transfered
// to new positions, velocities + kicks transfered to  newvels
// note if working in helio/bary coords we must not compute interaction
// terms from central mass, so this mass must be zeroed before calculation
// central mass is not allowed to change
// numBodiesMassive is the total number of massive bodies
// the total number of bodies is number of blocks times number of threads
// the number of bodies passed to the computeBodyaccel should be a
// multiple of pthreads as otherwise some shared mem is used but not zeroed
// numBodiesMassive should be an integer multiple of blocks
///////////////////////////////////////////////////////////

__global__ void interactionStep(double4* newPos, double4* newVel, 
                double4* oldPos, double4* oldVel,
                double deltaTime, double damping, double soft2,
                int numBodiesMassive)
{
    extern __shared__ double4 sharedPos[];
    int index = blockIdx.x*blockDim.x + threadIdx.x;

    double4 vel = oldVel[index]; 
    double4 pos = oldPos[index];    // mass transfered 


// compute acceleration with n^2 force step
    double3 accel = computeBodyAccel(pos, oldPos, numBodiesMassive, soft2); 

/* to do massless particles: 
   second argument here (oldPos) contains the massive particles,  
   with third argument specifying the number of massive particles 
   the first argument can contain massless particles with 
     index > numBodiesMassive.
   The computeBodyAccel does not use the blockIdx
   The total number of bodies numBodiestotal = blockDim.x*gridDim.x 
   which can be larger than numBodiesMassive 
   Then all particles with index > numBodiesMassive will not be included 
   in force computation
*/

/* This routine is the only one that knows the index of pos
    It is possible to vary this code so interactions between all planetesimals
    are not computed
    You would set the third argument in the above call to computeBodyAccel
    to be all bodies if the blockIdx.x is 0 (first 256=pthreads or so)
    and set it to only be blockDim.x if blockIDx.x >0
    This would allow 256 particles to interact with each other and feel
    forces from all other particles
    all particles would feel force from the 256
    the rest of the particles would not feel forces from each other
*/

 
   if (index ==0){ // don't ever kick the central mass
      accel.x = 0.0; accel.y = 0.0; accel.z = 0.0; 
   }
   // update velocity 
   vel.x += accel.x * deltaTime;
   vel.y += accel.y * deltaTime;
   vel.z += accel.z * deltaTime;  
 
// should be  1.0 if no damping 
//    vel.x *= damping;  vel.y *= damping;  vel.z *= damping; 
 
    // store position and new velocity 
    newPos[index] = pos; 
    newVel[index] = vel; 
    __syncthreads();
} 
 

#endif // #ifndef _NBODY_KERNEL_H_

