
#ifndef _SWEEP_KERNEL_H_
#define _SWEEP_KERNEL_H_

// #include "doub_all.h"


// close approach/encounter detection kernels
// all pairs sweeps
// in single precision
// note some use of sqrtf and rsqrtf

// return radius
__device__ float radius(float4 pos)
{
  return sqrtf(pos.x*pos.x + pos.y*pos.y + pos.z*pos.z);
}

// convert a double4 into a float4
__device__ float4 d4tof4(double4 dpos)
{
   float4 fpos;
   fpos.x = (float)dpos.x;
   fpos.y = (float)dpos.y;
   fpos.z = (float)dpos.z;
   fpos.w = (float)dpos.w;
   return fpos;
}

#define THIRDTOTHIRD 0.693361
// this number is (1/3)^(1/3)
#define THIRDTOTHIRDOVER2 0.346680637175317
// this number is (1/3)^(1/3)/2

/////////////////////////////////////
// return mutual hill radius of particle0,1
// rH mutual = [(m_i + m_j)/(3.0*m0)]^(1/3) * (a_i + a_j)/2.0
__device__ float rhillmutual(float4 pos0, float4 pos1, float m0)
{
   float r0 = radius(pos0);  // assumes position of sun at 0,0,0
   float r1 = radius(pos1);
   return pow((pos0.w + pos1.w)/m0,(float)0.33333333333333333)
         *(r0+r1)*THIRDTOTHIRDOVER2;
 
}

/////////////////////////////////////////////////////////////////
// returning semi major axis, eccentricity and escape velocity
__device__ void aev(float *a, float *e, float *vesc, 
          float4 Pos, float4 Vel, float GM)
{
   float r2 = Pos.x*Pos.x + Pos.y*Pos.y + Pos.z*Pos.z;  // 5 flops
   float recipr = rsqrtf(r2);  // 
   *vesc = sqrtf(2.0*GM*recipr);  // 3 flops

   float3 h;
   h.x = Pos.y*Vel.z - Pos.z*Vel.y;
   h.y = Pos.z*Vel.x - Pos.x*Vel.z;
   h.z = Pos.x*Vel.y - Pos.y*Vel.x; // 15 flops
   float h2 = h.x*h.x + h.y*h.y + h.z*h.z;  // 5 flops
   float v2 = Vel.x*Vel.x + Vel.y*Vel.y + Vel.z*Vel.z;  // 5 flops
   *a = 1.0/(2.0*recipr - v2);  // 3 flops
   *e = sqrtf(1.0 - h2/(*a));  // 3 flops ?
// total 31 flops about
}

// return only escape velocity and fast single precision
__device__ float getvesc(float4 Pos, float GM)
{
   float r2 = Pos.x*Pos.x + Pos.y*Pos.y + Pos.z*Pos.z;  // 5 flops
   float recipr = rsqrtf(r2);  // 1 flop
   return sqrtf(2.0*GM*recipr);  // 3 flops
}

///////////////////////////////////////////////////////
// First sweep, do a quick pass, only count pairs that can
// interact in a timestep
// count interactors for each particle
// inputs:
//       Pos,Vel     positions  and Vels 
//       deltaTime   timestep
//       numBodies   is used for tiling
//       numPlanets  chop here only look for close encounters to planets
//                        or massive bodies
//                   numPlanets here is actually meant numBodiesMassive
//       offset      to offset start of idi (not idj) if want to call multiple
//                   kernels because of mem requirements
//                   particles considered range from offset to numPlanets
//       GM          m0
//       m0          m0
// outputs:
//       npass    integer number of possible interactors
//                needs be numBodies big
//  numBodies should be a multiple of number of threads
// shared mem should be nthreads big
// npass should be zeroed before operation if you want to call this for
// fewer blocks than would cover all bodies x all bodies
//
// current strategy:
//   sweeping using escape velocity and distance between particles
//   note similarity between sweep1 and sweep2
//   we can store a pair twice as a second one can be removed when making 
//   encounter lists
///////////////////////////////////////////////////////
__global__ void sweep1(uint *npass, 
   double4* Pos, double4* Vel, double deltaTime, 
   int numBodies, int numPlanets, int offset, 
   double GM, double m0, double hillfac)
{
   extern __shared__ float4 sharedPos[];
   int p = blockDim.x;
   int idi = blockIdx.x*p + threadIdx.x + offset; 
   // index of the first particle
   float4 pos0 = d4tof4(Pos[idi]);
   float4 vel0 = d4tof4(Vel[idi]);
   float a0,e0,vesc;
   aev(&a0, &e0, &vesc, pos0, vel0,GM); // only order n
   float rh0 = hillfac*THIRDTOTHIRD*pow(2.0*pos0.w/m0,0.33333); 
         // times hillsphere  times 2 here for mutual corr
   int i,tile,np;
   np=0;
   // tile all particles for shared mem usage
   for (i = 0, tile = 0; i < numBodies; i += p, tile++){
      __syncthreads();
      // load up shared memory
      sharedPos[threadIdx.x] = d4tof4(Pos[i+threadIdx.x]); 
      __syncthreads(); // so all shared mem loaded
      if ((idi>0) && (idi<=numPlanets)){
        for(int j=0;j<p;j++){ // from here down order n^2
          int idj = i + j; // 1 operation
          if ((idj >0)&&(idi != idj)){
            // idj is the index of the second particle
            float4 pos1 = sharedPos[j]; // using shared mem for n^2 part
            float4 r; // 3 flops
            r.x = (pos0.x-pos1.x); r.y = (pos0.y-pos1.y); r.z = (pos0.z-pos1.z); 
            float dr = radius(r); // 6 flops
            if (dr< vesc*deltaTime + rh0) 
               np++; // number of possible approachers  // 1 operation
            }
        }
     }
   }
   if(idi>0) npass[idi-1] = np; //note offset, so scan doesn't need to pad by 1
   else npass[numBodies-1]=0; // central body not used 
                              // last index zeroed for scan
   __syncthreads();
}

///////////////////////////////////////////////////////
// Second sweep, same quick pass, 
// but now record the particle numbers
// of these pairs  in the array that is already the right size and
// with computed offsets using scan from output of sweep1
// inputs:
//       Pos,Vel     positions and velocities
//       deltaTime   timestep
//       npass_scan  is scanned number of interactors per particle
//       numBodies   used for tiling
//       numPlanets  chop interactions past some number of bodies
//       offset      to offset start of idi (not idj) 
//       GM          m0
//       m0          m0
// outputs:
//       pairlist  containing all interactor pairs 
// numBodies should be a multiple of number of threads
// shared mem should be nthreads big *float4
// pairlist does not have to be zeroed beforehand as every location used
// except last one if pairlis ist npairs+1 large 
// (I checked and a last zero is not needed as pairmark is scanned not pairlist)
///////////////////////////////////////////////////////
__global__ void sweep2(long long int* pairlist,
   double4* Pos, double4* Vel, double deltaTime, 
   uint *npass_scan, int numBodies, 
   int numPlanets, int offset, double GM, double m0, double hillfac)
{
   extern __shared__ float4 sharedPos[];
   int p = blockDim.x;
   int idi = blockIdx.x*p + threadIdx.x + offset;
   // idi is the index of the first particle
   int i,tile,np;
   float4 pos0 = d4tof4(Pos[idi]);
   float4 vel0 = d4tof4(Vel[idi]);
   float a0,e0,vesc;
   aev(&a0, &e0, &vesc, pos0, vel0,GM);
   float rh0 = hillfac*THIRDTOTHIRD*pow(2.0*pos0.w/m0,0.33333); 
         // times hillsphere  times 2 here for mutual corr
   np=0;
     // tile all particles
   for (i = 0, tile = 0; i < numBodies; i += p, tile++){
      __syncthreads();
      // load up shared memory
      sharedPos[threadIdx.x] = d4tof4(Pos[i+threadIdx.x]); 
      __syncthreads();
      if ((idi>0)&&(idi<= numPlanets)){
        for(int j=0;j<p;j++){ // could be unrolled, from here on order n^2
          int idj = i + j; // idj is the index of the second particle
          if ((idj >0) && (idi!=idj)){ //prevent sungrazers
             float4 pos1 = sharedPos[j];
             float4 r;
             r.x = (pos0.x-pos1.x); r.y = (pos0.y-pos1.y); r.z = (pos0.z-pos1.z);
             float dr = radius(r);
             if (dr< vesc*deltaTime + rh0){// 2 flops 
                long long int pairid = (long long int)idi*(long long int)numBodies+ (long long int) idj; // unique pairid
                uint plistid = npass_scan[idi-1] + np;
                pairlist[plistid] = pairid;
                np++; // number of possible approachers 
             }
          }
        }
     }
   }
   __syncthreads();
}

///////////////////////////////////////////////////////////
// Inputs: positions and velocities at beginning and end of timestep
//    for particle a and b
// Outputs 
//    Delta0, Delta1, dDelta0dt, dDelta1dt
//    distances between at beginning and end of timestep
//    and their time derivatives
///////////////////////////////////////////////////////////
__device__ void delta4(float4 posa0, float4 vela0, 
                  float4 posb0, float4 velb0,
                  float4 posa1, float4 vela1, 
                  float4 posb1, float4 velb1,
                  float* Delta0, float* Delta1, 
                  float* dDelta0dt, float* dDelta1dt, float deltaTime)
{
   float3 dr0,dr1,dv0,dv1;
   dr0.x=posa0.x-posb0.x; dr0.y=posa0.y-posb0.y; dr0.z=posa0.z-posb0.z;
   dr1.x=posa1.x-posb1.x; dr1.y=posa1.y-posb1.y; dr1.z=posa1.z-posb1.z;
   dv0.x=vela0.x-velb0.x; dv0.y=vela0.y-velb0.y; dv0.z=vela0.z-posb0.z;
   dv1.x=vela1.x-velb1.x; dv1.y=vela1.y-velb1.y; dv1.z=vela1.z-posb1.z;

// distances at beginning and end
   *Delta0 = sqrt(dr0.x*dr0.x + dr0.y*dr0.y + dr0.z*dr0.z);
   *Delta1 = sqrt(dr1.x*dr1.x + dr1.y*dr1.y + dr1.z*dr1.z);
// time derivative of distance at t=0 with extra factor deltaTime
   *dDelta0dt=  (dr0.x*dv0.x + dr0.y*dv0.y +
                        dr0.z*dv0.z) *deltaTime/(*Delta0);
// time derivative of distance at t=1 (multiplied by deltaTime)
   *dDelta1dt=  (dr1.x*dv1.x + dr1.y*dv1.y +
                        dr1.z*dv1.z) *deltaTime/(*Delta1);
}

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

////////////////////////////////////////////////////
/// evaluate equation 11 by Chambers 1999 for Delta
// dDeltadt assumed to be multiplied by tau
// The citation is Chambers, J. E. 1999, MNRAS, 304, 793-799,
// "A hybrid symplectic integrator that permits close encounters 
//  between massive bodies"
/////////////////////////////////////////////////////
__device__ float dist(float t, float Delta0, float Delta1,
        float dDelta0dt, float dDelta1dt)
{
   float u = 1.0-t;
   return u*u*(1.0 + 2.0*t)*Delta0 + t*t*(3.0-2.0*t)*Delta1
             + t*u*u*dDelta0dt - t*t*u*dDelta1dt;
// this is equation 11 Chambers et al. 99
}

/////////////////////////////////////////////////////
// find minimum distance using Chambers 99 procedures 
__device__ float  mindist(float Delta0, float Delta1, 
       float dDelta0dt, float dDelta1dt)
{
    float a = 6.0*(Delta0-Delta1)+ 3.0*(dDelta0dt + dDelta1dt);
    float b = 6.0*(Delta1-Delta0)- 2.0*(2.0*dDelta0dt + dDelta1dt);
    float c = dDelta0dt;
// these are from equations following 11 Chambers et al. 99
     // solve for t
    float x = (b*b - 4.0*a*c); // could be negative
    float mdist0 = MIN(Delta0,Delta1); // might want to keep track of this
    float mdist=-1.0;
    if ((x>0)&&(fabs(a)>0)) {
       float t1 = (-b + sqrt(x))/(2.0*a);
       float t2 = (-b - sqrt(x))/(2.0*a);
       float m1 = 1e5;
       float m2 = 1e5;  // something big
//       printf("t1,t2 %.2e %.2e\n",t1,t2);
       if ((t1>0.0) && (t1<1.0))
           m1 = dist(t1, Delta0, Delta1, dDelta0dt, dDelta1dt);
       if ((t2>0.0) && (t2<1.0))
           m2 = dist(t2, Delta0, Delta1, dDelta0dt, dDelta1dt);
       mdist = MIN(m1,m2);
     }
     mdist = MIN(mdist0,mdist);
     return mdist;
}

/////////////////////////////////////////////////////////////
// now sweeping using the pairlist not all possible pairs
// anything that approaches within hillfac times mutual hill spheres is judged
// a close encounter
//  Inputs:  
//     Pos0, Vel0  positions and velocities at beginning of timestep
//     Pos1, Vel1  positions and velocities at end of step
//     numBodies, npairs    numbers of Bodies and numbers of pairs in pairlist
//     pairlist    containing pairids
//     m0          central mass for hill sphere calcs
//     hillfac     factor to multiply mutual hill sphere
//  Outputs
//     pairlist    zero the pairid if there is no iteraction
//     pairmark    1 if interacting else zero
//                  should be length npairs+1
/////////////////////////////////////////////////////////////

__global__ void sweep3p(double4* Pos0, double4* Vel0, 
                        double4* Pos1, double4* Vel1,
                        double deltaTime, 
                        uint numBodies, uint npairs, 
                        long long int* pairlist, uint* pairmark,
                        double m0, double hillfac)
{
   int index = blockIdx.x*blockDim.x + threadIdx.x;
   if (index<npairs){ // may not be an integer mult of blockDim
     long long int pairid = pairlist[index];
     long long idi =  pairid/numBodies;
     long long idj =  pairid - idi*numBodies;
     float4 posa0,posa1,posb0,posb1,vela0,vela1,velb0,velb1;
     posa0 = d4tof4(Pos0[idi]); posa1 = d4tof4(Pos1[idi]);
     posb0 = d4tof4(Pos0[idj]); posb1 = d4tof4(Pos1[idj]);
     vela0 = d4tof4(Vel0[idi]); vela1 = d4tof4(Vel1[idi]);
     velb0 = d4tof4(Vel0[idj]); velb1 = d4tof4(Vel1[idj]);
     float Delta0, Delta1, dDelta0dt, dDelta1dt;
     delta4(posa0, vela0, posb0, velb0, posa1, vela1, posb1, velb1,
          &Delta0, &Delta1, &dDelta0dt, &dDelta1dt, 
         deltaTime);
     // assumes position of sun at 0,0,0
     float hill0 = rhillmutual(posa0,posb0,m0);
     float hill1 = rhillmutual(posa1,posb1,m0);
     float hill = hillfac*MIN(hill0,hill1);
     float md = mindist(Delta0,Delta1,dDelta0dt,dDelta1dt);
     if ((md > hill)||(idi*idj==0)){ // not interacting 
         pairlist[index]=0;  // zero if not interacting
         pairmark[index]=0;
     }
     else pairmark[index] = 1; // interacting
  }
  if (index==0) pairmark[npairs]=0; // set up for scan 

// if (index<=npairs)
//   printf("index=%d pm=%d pairid=%d\n",index,pairmark[index],pairlist[index]);
// note pairlist[npairs] is right now not used

   __syncthreads();
}

////////////////////////////////////////////
// repack a pairlist with fewer pairs using scan output
// we must use another array, can't be done with a single array
////////////////////////////////////////////
__global__ void repack(long long int* newpairlist, long long int* pairlist, 
                      uint* pairscan, int npairs, int newnpairs)
{
   int index = blockIdx.x*blockDim.x + threadIdx.x;
   if (index<npairs){
     if (pairscan[index+1] > pairscan[index]){
        newpairlist[pairscan[index]] = pairlist[index];
     }
   }
   __syncthreads();
}


#endif 


